require "socket"
require "log"
require "amq-protocol"
require "./channel_pool"
require "./client"
require "./upstream"

module AMQProxy
  class Server
    Log = ::Log.for(self)
    @clients_lock = Mutex.new
    @clients = Array(Client).new

    def initialize(upstream_host, upstream_port, upstream_tls, idle_connection_timeout = 5)
      tls_ctx = OpenSSL::SSL::Context::Client.new if upstream_tls
      @channel_pools = Hash(Credentials, ChannelPool).new do |hash, credentials|
        hash[credentials] = ChannelPool.new(upstream_host, upstream_port, tls_ctx, credentials, idle_connection_timeout)
      end
      Log.info { "Proxy upstream: #{upstream_host}:#{upstream_port} #{upstream_tls ? "TLS" : ""}" }
    end

    def client_connections
      @clients.size
    end

    def upstream_connections
      @channel_pools.each_value.sum &.connections
    end

    def listen(address, port)
      @server = server = TCPServer.new(address, port)
      Log.info { "Proxy listening on #{server.local_address}" }
      while socket = server.accept?
        begin
          addr = socket.remote_address
          spawn handle_connection(socket, addr), name: "Client#read_loop #{addr}"
        rescue IO::Error
          next
        end
      end
      Log.info { "Proxy stopping accepting connections" }
    end

    def stop_accepting_clients
      @server.try &.close
    end

    def disconnect_clients
      Log.info { "Disconnecting clients" }
      @clients_lock.synchronize do
        @clients.each &.close # send Connection#Close frames
      end
    end

    def close_sockets
      Log.info { "Closing client sockets" }
      @clients_lock.synchronize do
        @clients.each &.close_socket # close sockets forcefully
      end
    end

    private def handle_connection(socket, remote_address)
      c = Client.new(socket)
      active_client(c) do
        channel_pool = @channel_pools[c.credentials]
        c.read_loop(channel_pool)
      end
    rescue ex # only raise from constructor, when negotating
      Log.debug { "Client connection failure (#{remote_address}) #{ex.inspect}" }
    ensure
      socket.close rescue nil
    end

    private def active_client(client, &)
      @clients_lock.synchronize do
        @clients << client
      end
      yield client
    ensure
      @clients_lock.synchronize do
        @clients.delete client
      end
    end
  end
end
